/**
 * An object that serves as a validator.
 */
export declare type Validator<T> = {
    validate: (value: any) => ValidationResult<T>;
};
export declare type ValidationFailure = {
    isValid: false;
    error: IValidationError;
};
export declare type ValidationSuccess<T> = {
    value: T;
    isValid: true;
};
/**
 * An object that represents the result of a validation check.
 */
export declare type ValidationResult<T> = ValidationFailure | ValidationSuccess<T>;
/**
 * An object that represents a validation error.
 */
export declare type IValidationError = {
    readonly type: string;
    readonly errorMessage: string;
    readonly value: any;
} & {
    [key: string]: any;
};
/**
 * A base abstract class that represents a generic validation error.
 */
export declare abstract class ValidationError extends Error implements IValidationError {
    type: string;
    errorMessage: string;
    value: any;
    /**
     *
     * @param type A string representing what type of validation error that the
     *   error represents
     * @param errorMessage Some detail regarding the nature of the error
     * @param value The original value that triggered the validation error
     */
    constructor(type: string, errorMessage: string, value: any);
}
export declare class EitherError extends ValidationError {
    validationResults: ValidationResult<any>[];
    constructor(value: any, validationResults: ValidationResult<any>[]);
}
export declare class TupleError<T> extends ValidationError {
    validationResults: ValidationResult<T>[];
    constructor(value: any[], validationResults: ValidationResult<T>[]);
}
export declare class NotAnArrayError extends ValidationError {
    constructor(value: any);
}
export declare class UnexpectedArrayLengthError extends ValidationError {
    expectedLength: number;
    constructor(value: any[], expectedLength: number);
}
/**
 * A helper type for converting a Validator<T> type to a T
 *
 * For example, if you wanted to grab the validator from a `string()`, you'd use
 * this type like so:
 *
 * ```typescript
 * const stringValidator = string();
 *
 * type Str = InferType<typeof stringValidator>;
 * // Should be `type Str = string`
 * ```
 *
 * You can do this with objects as well. For example:
 *
 * ```typescript
 * const objectValidator = object({
 *   name: string(),
 *   email: string(),
 *   age: number()
 * });
 *
 * type Obj = InferType<typeof objectValidator>;
 * // Should be:
 * //
 * // type Obj = {
 * //   name: string
 * //   email: string
 * //   age: number
 * // }
 * ```
 */
export declare type InferType<V extends Validator<unknown>> = V extends Validator<infer T> ? T : never;
declare type ValidatorArrayToValidatorUnion<T extends Validator<unknown>[]> = T extends Validator<infer I>[] ? Validator<I> : never;
/**
 * A validator for validating objects against a list of validators.
 *
 * This is especially useful if a possible object has more than one possible
 * valid type.
 *
 * ## Usage
 *
 * ```typescript
 * const alts = alternatives(string(), number());
 *
 * const num = 10;
 * const str = "hello";
 * const bool = true;
 *
 * console.log(alts.validate(num).valid); // Should be true
 * console.log(alts.validate(str).valid); // Should be true
 * console.log(alts.validate(bool).valid); // Should be false
 * ```
 * @param alts A list of validators that are to be run
 * @returns A validator to validate an object against a set of validators
 */
export declare function either<T extends Validator<unknown>[]>(...alts: T): ValidatorArrayToValidatorUnion<T>;
declare type ValidatorTupleToValueTuple<T extends Validator<unknown>[]> = Validator<{
    [P in keyof T]: T[P] extends Validator<infer U> ? U : never;
}>;
/**
 * Used to validate a tuple against the individual values in an array.
 *
 * ## Usage
 *
 * ```typescript
 * const tup = tuple(string(), number());
 *
 * alts.validate(["a", 1]).isValid // will be true
 * alts.validate([1, "a"]).isValid // will be false
 * alts.validate([1]).isValid // will be false
 * ```
 * @param t The tuple of validators to validate a tuple against
 * @returns A validator to validate tuples
 */
export declare function tuple<T extends Validator<unknown>[]>(t: [...T]): ValidatorTupleToValueTuple<T>;
export declare class UnexpectedValueError extends ValidationError {
    constructor(value: any);
}
/**
 * Creates a validator for an object that rejects all values that passes the
 * invalidator.
 *
 * This validator is especially useful for cases where a value can be a string,
 * except for specific strings.
 *
 * For example:
 *
 * ```
 *const everythingButValidator = except(string(), exact("but"));
 *
 * everythingButValidator.validate("apples").isValid; // ✅
 * everythingButValidator.validate("bananas").isValid; // ✅
 * everythingButValidator.validate("cherries").isValid; // ✅
 * everythingButValidator.validate("but").isValid; // ❌
 * ```
 * @param validator The validator for which to validate the value against
 * @param invalidator The validator for which if is valid, the value will be
 *   rejected
 * @returns A Validator that will reject all values for which the invalidator
 *   validates the object
 */
export declare function except<T, I>(validator: Validator<T>, invalidator: Validator<I>): Validator<Exclude<T, I>>;
declare const _s: "string" | "number" | "bigint" | "boolean" | "symbol" | "undefined" | "object" | "function";
export declare type PossibleTypeof = typeof _s;
export declare class UnexpectedTypeofError extends ValidationError {
    expectedType: PossibleTypeof;
    constructor(value: any, expectedType: PossibleTypeof);
}
/**
 * Creates a validator that determines if the supplied value is a string.
 * @returns A validator to check if the value is of type string
 */
export declare const string: () => Validator<string>;
export declare type ExactTypes = string | number | boolean | null | undefined;
/**
 * Creates a validator that validates values that match the expected value
 * exactly.
 * @param expected The exact value to be expected
 * @returns A validator that will only validate values that match exactly the
 *   expected value
 */
export declare function exact<V extends ExactTypes>(expected: V): Validator<V>;
/**
 * Creates a validator that determines if the supplied value is a number.
 * @returns A validator to check if the value is of type number
 */
export declare const number: () => Validator<number>;
/**
 * Creates a validator that determines if the supplied value is a boolean.
 * @returns A validator to check if the value is of type boolean
 */
export declare const boolean: () => Validator<boolean>;
declare type BadValue<T> = {
    index: number;
    validation: ValidationResult<T>;
};
export declare class ArrayOfInvalidValuesError<T> extends ValidationError {
    badValues: BadValue<T>[];
    constructor(value: T[], errors: BadValue<T>[]);
}
/**
 * Creates a validator that determines if the supplied value is an array of the
 * specified validator.
 * @param validator The validator to validate the individual array values
 *   against
 * @returns A validator to check if the value is an array of the specified
 *   validator
 */
export declare function arrayOf<V>(validator: Validator<V>): Validator<V[]>;
export declare class BadObjectError extends ValidationError {
    faultyFields: {
        [key: string]: ValidationError;
    };
    constructor(value: any, faultyFields: {
        [key: string]: ValidationError;
    });
}
/**
 * Creates a validator that determines if the supplied value is an object, whose
 * fields contains are of nothing but types as defined by the specified
 * validator.
 * @param validator The validator to validate the individual fields in the
 *   object
 * @returns A validator that determines if the supplied value is an object,
 *   whose fields contains are of nothing but types as defined by the specified
 *   validator.
 */
export declare function objectOf<V>(validator: Validator<V>): Validator<{
    [keys: string]: V;
}>;
export declare class ValueIsUndefinedError extends ValidationError {
    constructor();
}
export declare class ValueIsNullError extends ValidationError {
    constructor();
}
/**
 * Creates a validator for an object, specified by the "schema".
 *
 * Each field in the "schema" is a validator, and each of them will validate
 * values against objects in concern.
 * @param schema An object containing fields of nothing but validators, each of
 *   which will be used to validate the value's respective fields
 * @returns A validator that will validate an object against the `schema`
 */
export declare function object<V extends object>(schema: {
    [key in keyof V]: Validator<V[key]>;
}): Validator<V>;
/**
 * Creates a validator that where the validation function will never determine
 * that a value is invalid
 * @returns A validator that will validate *all* objects
 */
export declare function any(): Validator<any>;
/**
 * Creates a validator that lazily evaluates the callback, at every validation.
 *
 * Useful for recursive types, such as a node for a tree.
 * @param schemaFn A function that returns a validator
 * @returns A validator, effectively just a "forwarding" of the validator
 *   returned by the `schemaFn`
 */
export declare function lazy<V extends any>(schemaFn: () => Validator<V>): Validator<V>;
export declare class TransformError extends ValidationError {
    errorObject: any;
    constructor(value: any, errorObject: any);
}
/**
 * Creates a validator that will parse the supplied value
 *
 * @param parse The parser function that will parse the supplied value
 * @returns A validator to validate the value against
 */
export declare const transform: <T>(parse: (value: any) => T) => Validator<T>;
/**
 * A validator creator that also accepts a predicate
 * @param validator The validator to run the predicate against
 * @param pred The predicate to run against the value
 * @returns A Validator, where if the predicate were to fail, it will result in
 *   a failed validation
 */
export declare function predicate<T>(validator: Validator<T>, pred: (value: T) => boolean): Validator<T>;
/**
 * A Validator creator that substitutes the error from one validator, to another
 * error for that validator.
 * @param validator The validator for which to have the error substituted
 * @param createError An error function that will return the appropriate error
 *   object
 * @returns A validator
 */
export declare function replaceError<T>(validator: Validator<T>, createError: (value: any, error: IValidationError) => IValidationError): Validator<T>;
/**
 * Chains two validators together.
 * @param left the first validator to validate values against
 * @param right the second validator to validate values against
 * @returns a validator that will validate first against the first validator
 *   then the second validator
 */
export declare const chain: <T1, T2>(left: Validator<T1>, right: Validator<T2>) => Validator<T2>;
/**
 * Gets the intersection of two validators
 * @param a The first validator to validate the object against
 * @param b The second validator to validate the object against
 * @returns A validator that is the intersection of the types represented by
 *   validators a and b
 */
export declare const intersection: <T1, T2>(a: Validator<T1>, b: Validator<T2>) => Validator<T1 & T2>;
/**
 * Creates a validator that can fallback to another value.
 * @param validator The validator that, if failed, will need a fallback
 * @param getFallback The function to acquire the fallback value
 * @returns A validator that should never be invalid
 */
export declare const fallback: <T1, T2>(validator: Validator<T1>, getFallback: () => T2) => Validator<T1 | T2>;
/**
 * This function validates and returns the parsed value. If validation failed,
 * it will throw a runtime exception
 * @param validator A validator to run the validation against the supplied value
 * @param value The value to run the validation against
 * @returns The outcome of the validation
 */
export declare const validate: <T>(validator: Validator<T>, value: any) => T;
export {};
